"""
Gantry's public API.
"""

import platform
import random
import string
import sys
import time
from fnmatch import fnmatch
from pathlib import Path
from typing import List, Optional, Sequence, Tuple, Union

import rich
from beaker import (
    Beaker,
    BeakerCancelationCode,
    BeakerExperimentSpec,
    BeakerGroup,
    BeakerJob,
    BeakerJobPriority,
    BeakerRetrySpec,
    BeakerSortOrder,
    BeakerTask,
    BeakerTaskResources,
    BeakerTaskSpec,
    BeakerWorkload,
    BeakerWorkloadStatus,
)
from beaker.exceptions import (
    BeakerExperimentConflict,
    BeakerImageNotFound,
    BeakerSecretNotFound,
)
from rich import print, prompt
from rich.status import Status

from . import constants, util
from .aliases import PathOrStr
from .exceptions import *
from .git_utils import GitRepoState
from .util import print_stderr
from .version import VERSION

__all__ = ["GitRepoState", "launch_experiment", "follow_workload"]


def _wait_for_job_to_start(
    *,
    beaker: Beaker,
    job: BeakerJob,
    status: Status,
    start_time: float,
    timeout: int = 0,
    show_logs: bool = True,
) -> BeakerJob:
    # Pull events until job is running (or fails)...
    events = set()
    while not (job.status.HasField("finalized") or (show_logs and job.status.HasField("started"))):
        if timeout > 0 and (time.monotonic() - start_time) > timeout:
            raise BeakerJobTimeoutError(f"Timed out while waiting for job '{job.id}' to finish")

        for event in beaker.job.list_summarized_events(
            job, sort_order=BeakerSortOrder.descending, sort_field="latest_occurrence"
        ):
            event_hashable = (event.latest_occurrence.ToSeconds(), event.latest_message)
            if event_hashable not in events:
                status.update(f"[i]{event.latest_message}[/]")
                events.add(event_hashable)
                time.sleep(0.5)

        time.sleep(0.5)
        job = beaker.job.get(job.id)

    return job


def _job_preempted(job: BeakerJob) -> bool:
    return job.status.status == BeakerWorkloadStatus.canceled and job.status.canceled_code in (
        BeakerCancelationCode.system_preemption,
        BeakerCancelationCode.user_preemption,
    )


def _validate_args(args: Sequence[str]):
    if not args:
        raise ConfigurationError(
            "[ARGS]... are required! For example:\n$ gantry run -- python -c 'print(\"Hello, World!\")'"
        )

    try:
        arg_index = sys.argv.index("--")
    except ValueError:
        raise ConfigurationError("[ARGS]... are required and must all come after '--'")

    # NOTE: if a value was accidentally provided to a flag, like '--preemptible false', click will
    # surprisingly add that value to the args. So we do a check here for that situation.
    given_args = sys.argv[arg_index + 1 :]
    invalid_args = args[: -len(given_args)]
    if invalid_args:
        raise ConfigurationError(
            f"Invalid options, found extra arguments before the '--': "
            f"{', '.join([repr(s) for s in invalid_args])}.\n"
            "Hint: you might be trying to pass a value to a FLAG option.\n"
            "Try 'gantry run --help' for help."
        )


def launch_experiment(
    args: Sequence[str],
    name: Optional[str] = None,
    description: Optional[str] = None,
    task_name: str = "main",
    workspace: Optional[str] = None,
    group_name: Optional[str] = None,
    clusters: Optional[Sequence[str]] = None,
    gpu_types: Optional[Sequence[str]] = None,
    hostnames: Optional[Sequence[str]] = None,
    beaker_image: Optional[str] = None,
    docker_image: Optional[str] = None,
    cpus: Optional[float] = None,
    gpus: Optional[int] = None,
    memory: Optional[str] = None,
    shared_memory: Optional[str] = None,
    datasets: Optional[Sequence[str]] = None,
    gh_token_secret: str = constants.GITHUB_TOKEN_SECRET,
    ref: Optional[str] = None,
    branch: Optional[str] = None,
    conda: Optional[PathOrStr] = None,
    pip: Optional[PathOrStr] = None,
    venv: Optional[str] = None,
    env_vars: Optional[Sequence[str]] = None,
    env_secrets: Optional[Sequence[str]] = None,
    dataset_secrets: Optional[Sequence[str]] = None,
    timeout: int = 0,
    task_timeout: Optional[str] = None,
    show_logs: bool = True,
    allow_dirty: bool = False,
    dry_run: bool = False,
    yes: bool = False,
    save_spec: Optional[PathOrStr] = None,
    priority: Optional[str] = None,
    install: Optional[str] = None,
    no_python: bool = False,
    no_conda: bool = False,
    replicas: Optional[int] = None,
    leader_selection: bool = False,
    host_networking: bool = False,
    propagate_failure: Optional[bool] = None,
    propagate_preemption: Optional[bool] = None,
    synchronized_start_timeout: Optional[str] = None,
    mounts: Optional[Sequence[str]] = None,
    weka: Optional[str] = None,
    budget: Optional[str] = None,
    preemptible: Optional[bool] = None,
    retries: Optional[int] = None,
    results: str = constants.RESULTS_DIR,
    skip_tcpxo_setup: bool = False,
    python_version: Optional[str] = None,
):
    """
    Launch an experiment on Beaker. Same as the ``gantry run`` command.
    """

    _validate_args(args)

    if install:
        if no_python:
            raise ConfigurationError("--no-python and --install='...' are mutually exclusive.")
        if pip:
            raise ConfigurationError("--pip='...' and --install='...' are mutually exclusive.")

    if beaker_image is None and docker_image is None:
        beaker_image = constants.DEFAULT_IMAGE
    elif (beaker_image is None) == (docker_image is None):
        raise ConfigurationError(
            "Either --beaker-image or --docker-image must be specified, but not both."
        )

    task_resources = BeakerTaskResources(
        cpu_count=cpus, gpu_count=gpus, memory=memory, shared_memory=shared_memory
    )

    # Get git information.
    git_config = GitRepoState.from_env(ref=ref, branch=branch)

    # Validate repo state.
    if ref is None and not allow_dirty and git_config.is_dirty:
        raise DirtyRepoError("You have uncommitted changes! Use --allow-dirty to force.")

    # Initialize Beaker client and validate workspace.
    with util.init_client(workspace=workspace, yes=yes) as beaker:
        if beaker_image is not None and beaker_image != constants.DEFAULT_IMAGE:
            try:
                beaker_image = beaker.image.get(beaker_image).id
            except BeakerImageNotFound:
                raise ConfigurationError(f"Beaker image '{beaker_image}' not found")

        if budget is None and not beaker.workspace.get().budget_id:
            budget = prompt.Prompt.ask(
                "[yellow]Missing '--budget' option, "
                "see https://beaker-docs.apps.allenai.org/concept/budgets.html for more information.[/]\n"
                "[i]Please enter the budget account to associate with this experiment[/]",
            )

            if not budget:
                raise ConfigurationError("Budget account must be specified!")

        # Maybe resolve or create group.
        group: Optional[BeakerGroup] = None
        if group_name is not None:
            group = util.resolve_group(beaker, group_name)
            if group is None:
                if prompt.Confirm.ask(
                    f"Group [green]{group_name}[/] not found in workspace, would you like to create this group?"
                ):
                    group = beaker.group.create(group_name)
                    print(f"Group created: {util.group_url(beaker, group)}")
                else:
                    print_stderr("[yellow]canceled[/]")
                    sys.exit(1)

        # Get the entrypoint dataset.
        entrypoint_dataset = util.ensure_entrypoint_dataset(beaker)

        # Get / set the GitHub token secret.
        if not git_config.is_public:
            try:
                beaker.secret.get(gh_token_secret)
            except BeakerSecretNotFound:
                print_stderr(
                    f"[yellow]GitHub token secret '{gh_token_secret}' not found in workspace.[/]\n"
                    f"You can create a suitable GitHub token by going to https://github.com/settings/tokens/new "
                    f"and generating a token with the '\N{ballot box with check} repo' scope."
                )
                gh_token = prompt.Prompt.ask(
                    "[i]Please paste your GitHub token here[/]",
                    password=True,
                )
                if not gh_token:
                    raise ConfigurationError("token cannot be empty!")
                beaker.secret.write(gh_token_secret, gh_token)
                print(
                    f"GitHub token secret uploaded to workspace as '{gh_token_secret}'.\n"
                    f"If you need to update this secret in the future, use the command:\n"
                    f"[i]$ gantry config set-gh-token[/]"
                )

            gh_token_secret = util.ensure_github_token_secret(beaker, gh_token_secret)

        # Validate the input datasets.
        datasets_to_use = ensure_datasets(beaker, *datasets) if datasets else []

        env_vars_to_use = []
        for e in env_vars or []:
            try:
                env_name, val = e.split("=", 1)
            except ValueError:
                raise ValueError("Invalid --env option: {e}")
            env_vars_to_use.append((env_name, val))

        env_secrets_to_use = []
        for e in env_secrets or []:
            try:
                env_secret_name, secret = e.split("=", 1)
            except ValueError:
                raise ValueError(f"Invalid --env-secret option: '{e}'")
            env_secrets_to_use.append((env_secret_name, secret))

        dataset_secrets_to_use = []
        for ds in dataset_secrets or []:
            try:
                secret, mount_path = ds.split(":", 1)
            except ValueError:
                raise ValueError(f"Invalid --dataset-secret option: '{ds}'")
            dataset_secrets_to_use.append((secret, mount_path))

        mounts_to_use = []
        for m in mounts or []:
            try:
                source, target = m.split(":", 1)
            except ValueError:
                raise ValueError(f"Invalid --mount option: '{m}'")
            mounts_to_use.append((source, target))

        weka_buckets = []
        for m in weka or []:
            try:
                source, target = m.split(":", 1)
            except ValueError:
                raise ValueError(f"Invalid --weka option: '{m}'")
            weka_buckets.append((source, target))

        # Validate clusters.
        if clusters or gpu_types:
            cl_objects = list(beaker.cluster.list())

            final_clusters = []
            for pat in clusters or ["*"]:
                org = beaker.org_name

                og_pat = pat
                if "/" in pat:
                    org, pat = pat.split("/", 1)

                matching_clusters = []
                for cl in cl_objects:
                    if not fnmatch(cl.name, pat) or cl.organization_name != org:
                        continue

                    if gpu_types:
                        cl_gpu_type = util.get_gpu_type(beaker, cl)
                        if not cl_gpu_type:
                            continue

                        for pattern in gpu_types:
                            if pattern.lower() in cl_gpu_type.lower():
                                break
                        else:
                            continue

                    matching_clusters.append(f"{cl.organization_name}/{cl.name}")

                if matching_clusters:
                    final_clusters.extend(matching_clusters)
                elif clusters:
                    raise ConfigurationError(
                        f"cluster '{og_pat}' did not match any Beaker clusters"
                    )
                elif gpu_types:
                    raise ConfigurationError(
                        f"""GPU type specs "{'", "'.join(gpu_types)}" did not match any Beaker clusters"""
                    )

            clusters = list(set(final_clusters))  # type: ignore

        # Default to preemptible when no cluster has been specified.
        if not clusters and preemptible is None:
            preemptible = True

        # Initialize experiment and task spec.
        spec = _build_experiment_spec(
            task_name=task_name,
            clusters=list(clusters or []),
            task_resources=task_resources,
            arguments=list(args),
            entrypoint_dataset=entrypoint_dataset.id,
            git_config=git_config,
            budget=budget,
            description=description,
            beaker_image=beaker_image,
            docker_image=docker_image,
            gh_token_secret=gh_token_secret if not git_config.is_public else None,
            conda=conda,
            pip=pip,
            venv=venv,
            datasets=datasets_to_use,
            env=env_vars_to_use,
            env_secrets=env_secrets_to_use,
            dataset_secrets=dataset_secrets_to_use,
            priority=priority,
            install=install,
            no_python=no_python,
            no_conda=no_conda,
            replicas=replicas,
            leader_selection=leader_selection,
            host_networking=host_networking or (bool(replicas) and leader_selection),
            propagate_failure=propagate_failure,
            propagate_preemption=propagate_preemption,
            synchronized_start_timeout=synchronized_start_timeout,
            task_timeout=task_timeout,
            mounts=mounts_to_use,
            weka_buckets=weka_buckets,
            hostnames=None if hostnames is None else list(hostnames),
            preemptible=preemptible,
            retries=retries,
            results=results,
            skip_tcpxo_setup=skip_tcpxo_setup,
            python_version=python_version,
        )

        if save_spec:
            if (
                Path(save_spec).is_file()
                and not yes
                and not prompt.Confirm.ask(
                    f"[yellow]The file '{save_spec}' already exists. "
                    f"[i]Are you sure you want to overwrite it?[/][/]"
                )
            ):
                raise KeyboardInterrupt
            spec.to_file(save_spec)
            print(f"Experiment spec saved to {save_spec}")

        if not name:
            default_name = util.unique_name()
            if yes:
                name = default_name
            else:
                name = prompt.Prompt.ask(
                    "[i]What would you like to call this experiment?[/]", default=util.unique_name()
                )

        if not name:
            raise ConfigurationError("Experiment name cannot be empty!")

        if dry_run:
            rich.get_console().rule("[b]Dry run[/]")
            print(
                f"[b]Workspace:[/] {beaker.workspace.url()}\n"
                f"[b]Group:[/] {None if group is None else util.group_url(beaker, group)}\n"
                f"[b]Commit:[/] {git_config.ref_url}\n"
                f"[b]Branch:[/] {git_config.branch_url}\n"
                f"[b]Name:[/] {name}\n"
                f"[b]Experiment spec:[/]",
                spec.to_json(),
            )
            return

        name_prefix = name
        while True:
            try:
                workload = beaker.experiment.create(name=name, spec=spec)
                break
            except BeakerExperimentConflict:
                name = (
                    name_prefix
                    + "-"
                    + random.choice(string.ascii_lowercase)
                    + random.choice(string.ascii_lowercase)
                    + random.choice(string.digits)
                    + random.choice(string.digits)
                )

        print(
            f"Experiment '{beaker.user_name}/{workload.experiment.name}' ({workload.experiment.id}) submitted.\n"
            f"Experiment URL: {beaker.workload.url(workload)}"
        )

        if group is not None:
            beaker.group.update(group, add_experiment_ids=[workload.experiment.id])
            print(f"Group URL: {util.group_url(beaker, group)}")

        # Can return right away if timeout is 0.
        if timeout == 0:
            return

        job: Optional[BeakerJob] = None
        try:
            job = follow_workload(beaker, workload, timeout=timeout, show_logs=show_logs)
        except (TermInterrupt, BeakerJobTimeoutError) as exc:
            print_stderr(f"[red][bold]{exc.__class__.__name__}:[/] [i]{exc}[/][/]")
            beaker.workload.cancel(workload)
            print_stderr("[yellow]Experiment cancelled.[/]")
            sys.exit(1)
        except KeyboardInterrupt:
            print_stderr("[yellow]Caught keyboard interrupt...[/]")
            if prompt.Confirm.ask("Would you like to cancel the experiment?"):
                beaker.workload.cancel(workload)
                print_stderr(f"[red]Experiment stopped:[/] {beaker.workload.url(workload)}")
                return
            else:
                print(f"See the experiment at {beaker.workload.url(workload)}")
                print_stderr(
                    f"[yellow]To cancel the experiment manually, run:\n[i]$ gantry stop {workload.experiment.id}[/][/]"
                )
                sys.exit(1)

        util.display_results(beaker, workload, job)


def _build_experiment_spec(
    *,
    task_name: str,
    clusters: List[str],
    task_resources: BeakerTaskResources,
    arguments: List[str],
    entrypoint_dataset: str,
    git_config: GitRepoState,
    budget: Optional[str] = None,
    description: Optional[str] = None,
    beaker_image: Optional[str] = None,
    docker_image: Optional[str] = None,
    gh_token_secret: Optional[str] = constants.GITHUB_TOKEN_SECRET,
    conda: Optional[PathOrStr] = None,
    pip: Optional[PathOrStr] = None,
    venv: Optional[str] = None,
    datasets: Optional[List[Tuple[str, Optional[str], str]]] = None,
    env: Optional[List[Tuple[str, str]]] = None,
    env_secrets: Optional[List[Tuple[str, str]]] = None,
    dataset_secrets: Optional[List[Tuple[str, str]]] = None,
    priority: Optional[Union[str, BeakerJobPriority]] = None,
    install: Optional[str] = None,
    no_python: bool = False,
    no_conda: bool = False,
    replicas: Optional[int] = None,
    leader_selection: bool = False,
    host_networking: bool = False,
    propagate_failure: Optional[bool] = None,
    propagate_preemption: Optional[bool] = None,
    synchronized_start_timeout: Optional[str] = None,
    task_timeout: Optional[str] = None,
    mounts: Optional[List[Tuple[str, str]]] = None,
    weka_buckets: Optional[List[Tuple[str, str]]] = None,
    hostnames: Optional[List[str]] = None,
    preemptible: Optional[bool] = None,
    retries: Optional[int] = None,
    results: str = constants.RESULTS_DIR,
    skip_tcpxo_setup: bool = False,
    python_version: Optional[str] = None,
):
    task_spec = (
        BeakerTaskSpec.new(
            task_name,
            beaker_image=beaker_image,
            docker_image=docker_image,
            result_path=results,
            command=["bash", "/gantry/entrypoint.sh"],
            arguments=arguments,
            resources=task_resources,
            priority=priority,
            preemptible=preemptible,
            replicas=replicas,
            leader_selection=leader_selection,
            host_networking=host_networking,
            propagate_failure=propagate_failure,
            propagate_preemption=propagate_preemption,
            synchronized_start_timeout=synchronized_start_timeout,
            timeout=task_timeout,
        )
        .with_env_var(name="GANTRY_VERSION", value=VERSION)
        .with_env_var(name="GITHUB_REPO", value=git_config.repo)
        .with_env_var(name="GIT_REF", value=git_config.ref)
        .with_env_var(name="GANTRY_TASK_NAME", value=task_name)
        .with_env_var(name="RESULTS_DIR", value=results)
        .with_dataset("/gantry", beaker=entrypoint_dataset)
    )

    if git_config.branch is not None:
        task_spec = task_spec.with_env_var(name="GIT_BRANCH", value=git_config.branch)

    if clusters:
        task_spec = task_spec.with_constraint(cluster=clusters)

    if hostnames:
        task_spec = task_spec.with_constraint(hostname=hostnames)

    if gh_token_secret is not None:
        task_spec = task_spec.with_env_var(name="GITHUB_TOKEN", secret=gh_token_secret)

    if skip_tcpxo_setup:
        task_spec = task_spec.with_env_var(name="GANTRY_SKIP_TCPXO_SETUP", value="1")

    for name, val in env or []:
        task_spec = task_spec.with_env_var(name=name, value=val)

    for name, secret in env_secrets or []:
        task_spec = task_spec.with_env_var(name=name, secret=secret)

    if no_python:
        task_spec = task_spec.with_env_var(name="NO_PYTHON", value="1")
    else:
        if not no_conda:
            if conda is not None:
                task_spec = task_spec.with_env_var(
                    name="CONDA_ENV_FILE",
                    value=str(conda),
                )
            elif Path(constants.CONDA_ENV_FILE).is_file():
                task_spec = task_spec.with_env_var(
                    name="CONDA_ENV_FILE",
                    value=constants.CONDA_ENV_FILE,
                )
            elif Path(constants.CONDA_ENV_FILE_ALTERNATE).is_file():
                task_spec = task_spec.with_env_var(
                    name="CONDA_ENV_FILE",
                    value=constants.CONDA_ENV_FILE_ALTERNATE,
                )
            else:
                task_spec = task_spec.with_env_var(
                    name="PYTHON_VERSION",
                    value=python_version
                    if python_version is not None
                    else ".".join(platform.python_version_tuple()[:-1]),
                )

            if venv is not None:
                task_spec = task_spec.with_env_var(
                    name="VENV_NAME",
                    value=venv,
                )
        else:
            task_spec = task_spec.with_env_var(name="NO_CONDA", value="1")

        if pip is not None:
            task_spec = task_spec.with_env_var(
                name="PIP_REQUIREMENTS_FILE",
                value=str(pip),
            )

        if install is not None:
            task_spec = task_spec.with_env_var(name="INSTALL_CMD", value=install)

    if datasets:
        for dataset_id, sub_path, path in datasets:
            task_spec = task_spec.with_dataset(path, beaker=dataset_id, sub_path=sub_path)

    for secret, mount_path in dataset_secrets or []:
        task_spec = task_spec.with_dataset(mount_path, secret=secret)

    if mounts:
        for source, target in mounts:
            task_spec = task_spec.with_dataset(target, host_path=source)

    if weka_buckets:
        for source, target in weka_buckets:
            task_spec = task_spec.with_dataset(target, weka=source)

    return BeakerExperimentSpec(
        description=description,
        budget=budget,
        tasks=[task_spec],
        retry=None if not retries else BeakerRetrySpec(allowed_task_retries=retries),
    )


def ensure_datasets(beaker: Beaker, *datasets: str) -> List[Tuple[str, Optional[str], str]]:
    out = []
    for dataset_str in datasets:
        dataset_name: str
        path: str
        sub_path: Optional[str] = None
        if dataset_str.count(":") == 1:
            dataset_name, path = dataset_str.split(":")
        elif dataset_str.count(":") == 2:
            dataset_name, sub_path, path = dataset_str.split(":")
        else:
            raise ValueError(
                f"Bad '--dataset' specification: '{dataset_str}'\n"
                f"Datasets should be in the form of 'dataset-name:/mount/location'"
                f"or 'dataset-name:sub/path:/mount/location'"
            )
        dataset_id = beaker.dataset.get(dataset_name).id
        out.append((dataset_id, sub_path, path))
    return out


def follow_workload(
    beaker: Beaker,
    workload: BeakerWorkload,
    *,
    task: BeakerTask | None = None,
    timeout: int = 0,
    tail: bool = False,
    show_logs: bool = True,
) -> BeakerJob:
    """
    Follow a workload until completion while streaming logs to stdout.

    :param task: A specific task in the workload to follow. Defaults to the first task.
    :param timeout: The number of seconds to wait for the workload to complete. Raises a timeout
        error if it doesn't complete in time. Set to 0 (the default) to wait indefinitely.
    :param tail: Start tailing the logs if a job is already running. Otherwise shows all logs.
    :param show_logs: Set to ``False`` to avoid streaming the logs.

    :returns: The finalized :class:`~beaker.types.BeakerJob` from the task being followed.

    :raises ~gantry.exceptions.BeakerJobTimeoutError: If ``timeout`` is set to a positive number
        and the workload doesn't complete in time.
    """
    console = rich.get_console()
    start_time = time.monotonic()
    preempted_job_ids = set()

    while True:
        with console.status("[i]waiting...[/]", spinner="point", speed=0.8) as status:
            # Wait for job to be created...
            job: BeakerJob | None = None
            while job is None:
                if (
                    j := beaker.workload.get_latest_job(workload, task=task)
                ) is not None and j.id not in preempted_job_ids:
                    job = j
                else:
                    time.sleep(1.0)

            # Wait for job to start...
            job = _wait_for_job_to_start(
                beaker=beaker,
                job=job,
                status=status,
                start_time=start_time,
                timeout=timeout,
                show_logs=show_logs,
            )

        # Stream logs...
        if show_logs and job.status.HasField("started"):
            print()
            rich.get_console().rule("Logs")

            for job_log in beaker.job.logs(job, tail_lines=10 if tail else None, follow=True):
                console.print(job_log.message.decode(), highlight=False, markup=False)
                if timeout > 0 and (time.monotonic() - start_time) > timeout:
                    raise BeakerJobTimeoutError(
                        f"Timed out while waiting for job '{job.id}' to finish"
                    )

            print()
            rich.get_console().rule("End logs")
            print()

        # Wait for job to finalize...
        while not job.status.HasField("finalized"):
            time.sleep(0.5)
            job = beaker.job.get(job.id)

        # If job was preempted, we start over...
        if _job_preempted(job):
            print(f"[yellow]Job '{job.id}' preempted.[/] ")
            preempted_job_ids.add(job.id)
            continue

        return job



from typing import Optional

import click
from beaker import BeakerWorkload
from rich import print

from .. import util
from ..api import follow_workload
from ..exceptions import ConfigurationError, NotFoundError
from .main import CLICK_COMMAND_DEFAULTS, main


@main.command(**CLICK_COMMAND_DEFAULTS)
@click.argument("workload", nargs=1, required=False, type=str)
@click.option(
    "-t", "--tail", is_flag=True, help="Only tail the logs as opposed to printing all logs so far."
)
@click.option(
    "-l",
    "--latest",
    is_flag=True,
    help="Get the logs from the latest running experiment (non-session) workload.",
)
@click.option(
    "-w",
    "--workspace",
    type=str,
    help="""The Beaker workspace to pull experiments from.
    If not specified, your default workspace will be used.""",
)
@click.option(
    "-a",
    "--author",
    type=str,
    help="Pull the latest experiment workload for a particular author. Defaults to your own account.",
)
def follow(
    workload: Optional[str] = None,
    tail: bool = False,
    latest: bool = False,
    workspace: Optional[str] = None,
    author: Optional[str] = None,
):
    """
    Follow the logs for a running experiment.
    """
    with util.init_client(ensure_workspace=False) as beaker:
        wl: Optional[BeakerWorkload] = None
        if workload is not None:
            if latest or workspace is not None or author is not None:
                raise ConfigurationError(
                    "[WORKLOAD] is mutually exclusive with -a/--author, -w/--workspace, and -l/--latest"
                )
            wl = beaker.workload.get(workload)
        else:
            if not latest:
                raise ConfigurationError(
                    "A filter such as -l/--latest is required when no [WORKLOAD] is specified"
                )

            wl = util.get_latest_workload(
                beaker, author_name=author, workspace_name=workspace, running=True
            )
            if wl is None:
                raise NotFoundError("Failed to find an experiment workload to follow")

            print(
                f"Following experiment [b cyan]{wl.experiment.name}[/] ({wl.experiment.id}) at {beaker.workload.url(wl)}"
            )

        assert wl is not None
        job = follow_workload(beaker, wl, tail=tail)
        util.display_results(beaker, wl, job)



from typing import List, Optional, Sequence

import click
from beaker import BeakerWorkload
from beaker.exceptions import BeakerWorkloadNotFound
from rich import print, prompt

from .. import util
from ..exceptions import ConfigurationError, NotFoundError
from .main import CLICK_COMMAND_DEFAULTS, main


@main.command(**CLICK_COMMAND_DEFAULTS)
@click.argument("workload", nargs=-1, type=str)
@click.option(
    "-l", "--latest", is_flag=True, help="""Stop your latest experiment (non-session) workload."""
)
@click.option(
    "-w",
    "--workspace",
    type=str,
    help="""The Beaker workspace to pull experiments from.
    If not specified, your default workspace will be used.""",
)
@click.option("--dry-run", is_flag=True, help="Do a dry-run without stopping any experiments.")
@click.option(
    "-y",
    "--yes",
    is_flag=True,
    help="""Skip all confirmation prompts.""",
)
def stop(
    workload: Sequence[str],
    latest: bool = False,
    workspace: Optional[str] = None,
    dry_run: bool = False,
    yes: bool = False,
):
    """
    Stop a running workload.
    """
    if workload and latest:
        raise ConfigurationError("-l/--latest is mutually exclusive with [WORKLOAD] args")

    beaker = util.init_client(ensure_workspace=False)

    workloads: List[BeakerWorkload] = []
    if workload:
        for workload_name in workload:
            try:
                workloads.append(beaker.workload.get(workload_name))
            except BeakerWorkloadNotFound:
                raise NotFoundError(f"Workload '{workload_name}' not found")
    elif latest:
        wl = util.get_latest_workload(beaker, workspace_name=workspace, running=True)
        if wl is None:
            print("[yellow]No running workloads to stop[/]")
        else:
            workloads.append(wl)

    for wl in workloads:
        if dry_run:
            print(f"[b yellow]Dry run:[/] would stop [b cyan]{wl.experiment.name}[/]")
        else:
            if not yes and not prompt.Confirm.ask(
                f"Stop experiment [b cyan]{wl.experiment.name}[/] at [blue u]{beaker.workload.url(wl)}[/]?"
            ):
                print("[yellow]Skipping experiment...[/]")
                continue

            try:
                beaker.workload.cancel(wl)
            except (BeakerWorkloadNotFound,):
                pass
            print(
                f"[b green]\N{check mark}[/] [b cyan]{wl.experiment.name}[/] at {beaker.workload.url(wl)} stopped"
            )



import os

from .main import CLICK_COMMAND_DEFAULTS, CLICK_GROUP_DEFAULTS, main


@main.group(**CLICK_GROUP_DEFAULTS)
def completion():
    """
    Generate the autocompletion script for gantry for the specified shell.

    See each sub-command's help for details on how to use the generated script.
    """


@completion.command(**CLICK_COMMAND_DEFAULTS)
def bash():
    """
    Generate the autocompletion script for bash.

    Save the output somewhere and then source the file:

    $ gantry --quiet completion bash > ~/.gantry-complete.bash && . ~/.gantry-complete.bash
    """
    os.environ["_GANTRY_COMPLETE"] = "bash_source"
    main()


@completion.command(**CLICK_COMMAND_DEFAULTS)
def fish():
    """
    Generate the autocompletion script for fish.

    Execute this once to setup completions:

    $ gantry --quiet completion fish > ~/.config/fish/completions/gantry.fish
    """
    os.environ["_GANTRY_COMPLETE"] = "fish_source"
    main()


@completion.command(**CLICK_COMMAND_DEFAULTS)
def zsh():
    """
    Generate the autocompletion script for zsh.

    Save the output somewhere and then source the file:

    $ gantry --quiet completion bash > ~/.gantry-complete.zsh && . ~/.gantry-complete.zsh
    """
    os.environ["_GANTRY_COMPLETE"] = "zsh_source"
    main()



from typing import Optional

import click
from beaker import Beaker, BeakerJob, BeakerTask, BeakerWorkload
from rich import print

from .. import util
from ..exceptions import ConfigurationError
from .main import CLICK_COMMAND_DEFAULTS, main


def _get_job(
    beaker: Beaker, wl: BeakerWorkload, task: BeakerTask, run: Optional[int] = None
) -> Optional[BeakerJob]:
    if run is None:
        return beaker.workload.get_latest_job(wl, task=task)
    else:
        # NOTE: ascending sort order on creation time is not implemented server-side yet
        jobs = list(reversed(list(beaker.job.list(task=task, sort_field="created"))))
        try:
            return jobs[run - 1]
        except IndexError:
            raise ConfigurationError(f"run number {run} does not exist")


@main.command(**CLICK_COMMAND_DEFAULTS)
@click.argument("workload", nargs=1, type=str)
@click.option("-r", "--replica", type=int, help="""The replica rank to pull logs from.""")
@click.option("--task", "task_name", type=str, help="""The name of task to pull logs from.""")
@click.option("-t", "--tail", type=int, help="""Tail this many lines.""")
@click.option("--run", type=int, help="""The run number to pull logs from.""")
@click.option(
    "-f", "--follow", is_flag=True, help="""Continue streaming logs for the duration of the job."""
)
def logs(
    workload: str,
    replica: Optional[int] = None,
    task_name: Optional[int] = None,
    tail: Optional[int] = None,
    run: Optional[int] = None,
    follow: bool = False,
):
    """
    Display the logs for an experiment workload.
    """
    if replica is not None and task_name is not None:
        raise ConfigurationError("--replica and --task are mutually exclusive")

    if run is not None and run < 1:
        raise ConfigurationError("--run must be at least 1")

    with util.init_client(ensure_workspace=False) as beaker:
        wl = beaker.workload.get(workload)
        tasks = list(wl.experiment.tasks)

        job: Optional[BeakerJob] = None
        task: Optional[BeakerTask] = None
        if replica is not None:
            for task in tasks:
                job = _get_job(beaker, wl, task, run=run)
                if job is not None and job.system_details.replica_group_details.rank == replica:
                    break
            else:
                raise ConfigurationError(f"Invalid replica rank '{replica}'")
        elif task_name is not None:
            for task in tasks:
                if task.name == task_name:
                    job = _get_job(beaker, wl, task, run=run)
                    break
            else:
                raise ConfigurationError(f"Invalid task name '{task_name}'")
        else:
            task = tasks[0]
            job = _get_job(beaker, wl, task, run=run)

        if job is None:
            print("[yellow]Experiment has not started yet[/]")
            return

        print(f"Showing logs from job '{job.id}' for task '{task.name}'...")
        util.display_logs(beaker, job, tail_lines=tail, follow=follow)



import logging
import os
import signal
import sys

import click
import rich
from beaker.exceptions import BeakerError
from click_help_colors import HelpColorsCommand, HelpColorsGroup
from click_option_group import optgroup
from rich import pretty, traceback
from rich.logging import RichHandler

from .. import util
from ..exceptions import *
from ..util import print_stderr
from ..version import VERSION

CLICK_GROUP_DEFAULTS = {
    "cls": HelpColorsGroup,
    "help_options_color": "green",
    "help_headers_color": "yellow",
    "context_settings": {"max_content_width": 115},
}

CLICK_COMMAND_DEFAULTS = {
    "cls": HelpColorsCommand,
    "help_options_color": "green",
    "help_headers_color": "yellow",
    "context_settings": {"max_content_width": 115},
}


def new_optgroup(name: str):
    return optgroup.group(f"\n ❯❯❯ {name}")


def excepthook(exctype, value, tb):
    """
    Used to patch `sys.excepthook` in order to customize handling of uncaught exceptions.
    """
    # Ignore in-house error types because we don't need a traceback for those.
    if issubclass(exctype, (GantryError, BeakerError)):
        print_stderr(f"[red][bold]{exctype.__name__}:[/] [i]{value}[/][/]")
    # For interruptions, call the original exception handler.
    elif issubclass(exctype, (KeyboardInterrupt, TermInterrupt)):
        sys.__excepthook__(exctype, value, tb)
    else:
        print_stderr(traceback.Traceback.from_exception(exctype, value, tb, suppress=[click]))


sys.excepthook = excepthook


def handle_sigterm(sig, frame):
    del sig, frame
    raise TermInterrupt


@click.group(**CLICK_GROUP_DEFAULTS)  # type: ignore[call-overload]
@click.version_option(version=VERSION)
@click.option(
    "--quiet",
    is_flag=True,
    help="Don't display the gantry logo.",
)
@click.option(
    "--log-level",
    type=click.Choice(["debug", "info", "warning", "error"]),
    show_choices=True,
    show_default=True,
    default="warning",
    help="The Python log level.",
)
def main(quiet: bool = False, log_level: str = "warning"):
    # Configure rich.
    if os.environ.get("GANTRY_GITHUB_TESTING"):
        # Force a broader terminal when running tests in GitHub Actions.
        console_width = 180
        rich.reconfigure(width=console_width, force_terminal=True, force_interactive=False)
        pretty.install()
    else:
        pretty.install()

    # Configure logging.
    logging.basicConfig(
        level=getattr(logging, log_level.upper()),
        format="%(message)s",
        handlers=[RichHandler(log_time_format="❯ [GANTRY (local)] [%X]")],
    )

    # Handle SIGTERM just like KeyboardInterrupt
    signal.signal(signal.SIGTERM, handle_sigterm)

    if not quiet:
        print_stderr(
            r'''
[cyan b]                                             o=======[]   [/]
[cyan b]   __ _                    _               _ |_      []   [/]
[cyan b]  / _` |  __ _    _ _     | |_      _ _   | || |     []   [/]
[cyan b]  \__, | / _` |  | ' \    |  _|    | '_|   \_, |   _/ ]_  [/]
[cyan b]  |___/  \__,_|  |_||_|   _\__|   _|_|_   _|__/   |_____| [/]
[blue b]_|"""""|_|"""""|_|"""""|_|"""""|_|"""""|_| """"| [/]
[blue b] `---------------------------------------------' [/]
''',  # noqa: W605
            highlight=False,
        )

    util.check_for_upgrades()



from typing import Optional

import click
from rich import print

from .. import constants, util
from .main import CLICK_COMMAND_DEFAULTS, CLICK_GROUP_DEFAULTS, main


@main.group(**CLICK_GROUP_DEFAULTS)
def config():
    """
    Configure Gantry for a specific Beaker workspace.
    """


@config.command(**CLICK_COMMAND_DEFAULTS)  # type: ignore
@click.argument("token")
@click.option(
    "-w",
    "--workspace",
    type=str,
    help="""The Beaker workspace to use.
    If not specified, your default workspace will be used.""",
)
@click.option(
    "-s",
    "--secret",
    type=str,
    help="""The name of the Beaker secret to write to.""",
    default=constants.GITHUB_TOKEN_SECRET,
    show_default=True,
)
@click.option(
    "-y",
    "--yes",
    is_flag=True,
    help="""Skip all confirmation prompts.""",
)
def set_gh_token(
    token: str,
    workspace: Optional[str] = None,
    secret: str = constants.GITHUB_TOKEN_SECRET,
    yes: bool = False,
):
    """
    Set or update Gantry's GitHub token for the workspace.

    You can create a suitable GitHub token by going to https://github.com/settings/tokens/new
    and generating a token with the '\N{ballot box with check} repo' scope.

    Example:

    $ gantry config set-gh-token "$GITHUB_TOKEN"
    """
    # Initialize Beaker client and validate workspace.
    with util.init_client(workspace=workspace, yes=yes) as beaker:
        # Write token to secret.
        beaker.secret.write(secret, token)

        print(
            f"[green]\N{check mark} GitHub token added to workspace "
            f"'{beaker.config.default_workspace}' as the secret '{secret}'"
        )



import click
import rich
from beaker import BeakerSortOrder
from rich import print
from rich.table import Table

from .. import util
from .main import CLICK_COMMAND_DEFAULTS, main


@main.command(name="find-gpus", **CLICK_COMMAND_DEFAULTS)
@click.option(
    "-a",
    "--all",
    "show_all",
    is_flag=True,
    help="Show all clusters, not just ones with free GPUs.",
)
@click.option(
    "-g",
    "--gpu-type",
    "gpu_types",
    type=str,
    multiple=True,
    help="""Filter by GPU type (e.g. "h100"). Multiple allowed.""",
)
def find_gpus_cmd(show_all: bool = False, gpu_types: tuple[str, ...] = tuple()):
    """
    Find free GPUs.
    """
    gpu_types = tuple(pattern.lower() for pattern in gpu_types)

    with util.init_client(ensure_workspace=False) as beaker:
        table = Table(title="Clusters", show_lines=True)
        table.add_column("Name", justify="left", no_wrap=True)
        table.add_column("Free GPUs", justify="left", no_wrap=True)
        table.add_column("GPU Type", justify="left", no_wrap=True)
        table.add_column("Slots", justify="left", no_wrap=True)

        with rich.get_console().status(
            "[i]collecting clusters...[/]", spinner="point", speed=0.8
        ) as status:
            for cluster in beaker.cluster.list(
                sort_field="free_gpus",
                sort_order=BeakerSortOrder.descending,
                include_cluster_occupancy=True,
            ):
                status.update(f"[i]collecting clusters...[/] {cluster.name}")

                if not show_all and cluster.cluster_occupancy.slot_counts.available == 0:
                    break

                gpu_type = util.get_gpu_type(beaker, cluster)
                if gpu_types:
                    if not gpu_type:
                        continue

                    match = gpu_type.lower()
                    for pattern in gpu_types:
                        if pattern in match:
                            gpu_type = util.highlight_pattern(gpu_type, pattern)
                            break
                    else:
                        continue

                gpus_available = cluster.cluster_occupancy.slot_counts.available
                gpus_available_style: str
                if gpus_available == 0:
                    gpus_available_style = "red"
                elif gpus_available < 8:
                    gpus_available_style = "yellow"
                else:
                    gpus_available_style = "green"

                table.add_row(
                    f"[b cyan]{beaker.org_name}/{cluster.name}[/]\n[u i blue]{beaker.cluster.url(cluster)}[/]",
                    f"[{gpus_available_style}]{gpus_available}[/]",
                    f"{gpu_type or 'UNKNOWN'}",
                    f"{cluster.cluster_occupancy.slot_counts.assigned}/{cluster.cluster_occupancy.slot_counts.total}",
                )

        print(table)



import concurrent.futures
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta, timezone
from itertools import islice
from typing import Generator, Iterable, List, Optional, Tuple

import click
import rich
from beaker import (
    Beaker,
    BeakerSortOrder,
    BeakerTask,
    BeakerWorkload,
    BeakerWorkloadStatus,
    BeakerWorkloadType,
)
from beaker.exceptions import BeakerGroupNotFound
from rich import print
from rich.table import Table

from .. import util
from ..exceptions import ConfigurationError
from .main import CLICK_COMMAND_DEFAULTS, main


class Defaults:
    limit = 10
    max_age = 7


@main.command(name="list", **CLICK_COMMAND_DEFAULTS)
@click.option(
    "-w",
    "--workspace",
    type=str,
    help="""The Beaker workspace to pull experiments from.
    If not specified, your default workspace will be used.""",
)
@click.option("-g", "--group", type=str, help="""The Beaker group to pull experiments from.""")
@click.option(
    "-l",
    "--limit",
    type=int,
    default=Defaults.limit,
    help=f"Limit the number of experiments to display. Default: {Defaults.limit}",
)
@click.option(
    "-a",
    "--author",
    type=str,
    help="Filter by author. Tip: use '--me' instead to show your own experiments.",
)
@click.option(
    "--me",
    is_flag=True,
    help="Only show your own experiments. Mutually exclusive with '--author'.",
)
@click.option(
    "-s",
    "--status",
    type=click.Choice([x.name for x in BeakerWorkloadStatus]),
    help="Filter by status. Multiple allowed.",
    multiple=True,
)
@click.option(
    "--max-age",
    type=int,
    default=Defaults.max_age,
    help=f"Maximum age of experiments, in days. Default: {Defaults.max_age}",
)
@click.option(
    "--all",
    "show_all",
    is_flag=True,
    help="Show all experiments, not just onces submitted through Gantry.",
)
def list_cmd(
    workspace: Optional[str] = None,
    group: Optional[str] = None,
    limit: int = Defaults.limit,
    author: Optional[str] = None,
    me: bool = False,
    status: Optional[List[str]] = None,
    max_age: int = Defaults.max_age,
    show_all: bool = False,
):
    """
    List recent experiments within a workspace or group.
    This will only show experiments launched with Gantry by default, unless '--all' is specified.
    """
    with util.init_client(ensure_workspace=False) as beaker:
        table = Table(title="Experiments", show_lines=True)
        table.add_column("Name", justify="left", no_wrap=True)
        table.add_column("Author", justify="left", style="blue", no_wrap=True)
        table.add_column("Created", justify="left", no_wrap=True)
        table.add_column("Tasks")

        if me:
            if author is not None:
                raise ConfigurationError("--me and -a/--author are mutually exclusive.")
            else:
                author = beaker.user_name

        status_msg = "[i]collecting experiments...[/]"
        with rich.get_console().status(status_msg, spinner="point", speed=0.8) as status_:
            with ThreadPoolExecutor() as executor:
                for wl, tasks in islice(
                    iter_workloads(
                        beaker,
                        executor=executor,
                        workspace=workspace,
                        group=group,
                        author=author,
                        statuses=status,
                        max_age=max_age,
                        show_all=show_all,
                        limit=None if not show_all else limit,
                    ),
                    limit,
                ):
                    status_.update(f"{status_msg} [cyan]{wl.experiment.name}[/]")

                    task_status_futures = []
                    for task in tasks:
                        task_status_futures.append(executor.submit(format_task, beaker, wl, task))

                    task_statuses = {}
                    for task_future in concurrent.futures.as_completed(task_status_futures):
                        task, task_status = task_future.result()
                        task_statuses[task.id] = task_status
                        status_.update(f"{status_msg} [cyan]{wl.experiment.name} ❯ [/]{task.name}")

                    table.add_row(
                        f"[b cyan]{wl.experiment.name}[/]\n[u i blue]{beaker.workload.url(wl)}[/]",
                        beaker.user.get(wl.experiment.author_id).name,
                        wl.experiment.created.ToDatetime(timezone.utc)
                        .astimezone(tz=None)
                        .strftime("%I:%M %p on %a, %b %-d"),
                        "\n".join(task_statuses[task.id] for task in tasks),
                    )

        print(table)


def is_gantry_workload(beaker: Beaker, wl: BeakerWorkload) -> bool:
    spec = beaker.experiment.get_spec(wl.experiment)
    for task_spec in spec.tasks:
        for env_var in task_spec.env_vars or []:
            if env_var.name == "GANTRY_VERSION":
                return True
    return False


def iter_workloads(
    beaker: Beaker,
    *,
    executor: ThreadPoolExecutor,
    workspace: Optional[str],
    group: Optional[str],
    author: Optional[str],
    statuses: Optional[List[str]],
    max_age: int,
    show_all: bool,
    limit: Optional[int] = None,
) -> Generator[Tuple[BeakerWorkload, Iterable[BeakerTask]], None, None]:
    created_after = datetime.now(tz=timezone.utc).astimezone() - timedelta(days=max_age)
    author_id = None if author is None else beaker.user.get(author)
    workload_statuses = (
        None if not statuses else [BeakerWorkloadStatus.from_any(s) for s in statuses]
    )

    if group is not None:
        beaker_group = util.resolve_group(
            beaker, group, workspace, fall_back_to_default_workspace=False
        )
        if beaker_group is None:
            raise BeakerGroupNotFound(group)

        # Will have to sort workloads manually by creation time, so we gather them all first.
        workload_futures = []
        for task_metrics in beaker.group.list_task_metrics(beaker_group):
            workload_futures.append(
                executor.submit(beaker.workload.get, task_metrics.experiment_id)
            )

        workloads = []
        for future in concurrent.futures.as_completed(workload_futures):
            wl = future.result()

            # Filter out non-gantry experiments.
            if not show_all and not is_gantry_workload(beaker, wl):
                continue

            if author_id is not None and wl.experiment.author_id != author_id:
                continue

            if workload_statuses is not None and wl.status not in workload_statuses:
                continue

            if wl.experiment.created.ToDatetime(timezone.utc) < created_after:
                continue

            workloads.append(wl)

        workloads.sort(key=lambda wl: wl.experiment.created.ToMilliseconds(), reverse=True)
        for wl in workloads:
            yield wl, wl.experiment.tasks
    else:
        for wl in beaker.workload.list(
            workspace=None if workspace is None else beaker.workspace.get(workspace),
            author=author_id,
            created_after=created_after,
            workload_type=BeakerWorkloadType.experiment,
            statuses=workload_statuses,
            sort_order=BeakerSortOrder.descending,
            sort_field="created",
            limit=limit,
        ):
            # Filter out non-gantry experiments.
            if not show_all and not is_gantry_workload(beaker, wl):
                continue

            yield wl, wl.experiment.tasks


def get_status(
    beaker: Beaker, wl: BeakerWorkload, task: BeakerTask
) -> Optional[BeakerWorkloadStatus]:
    job = beaker.workload.get_latest_job(wl, task=task)
    if job is not None:
        return BeakerWorkloadStatus.from_any(job.status.status)
    else:
        return None


def format_task(beaker: Beaker, wl: BeakerWorkload, task: BeakerTask) -> tuple[BeakerTask, str]:
    style = "i"
    status = get_status(beaker, wl, task)
    if status in (
        BeakerWorkloadStatus.running,
        BeakerWorkloadStatus.succeeded,
        BeakerWorkloadStatus.uploading_results,
        BeakerWorkloadStatus.ready_to_start,
        BeakerWorkloadStatus.initializing,
    ):
        style += " green"
    elif status in (
        BeakerWorkloadStatus.submitted,
        BeakerWorkloadStatus.queued,
        BeakerWorkloadStatus.canceled,
        BeakerWorkloadStatus.stopping,
    ):
        style += " yellow"
    elif status == BeakerWorkloadStatus.failed:
        style += " red"

    status_str = "unknown" if status is None else status.name

    return task, f"[i]{task.name}[/] ([{style}]{status_str}[/])"



from __future__ import annotations

import logging
from dataclasses import dataclass
from functools import cached_property
from typing import cast

import requests
from git import InvalidGitRepositoryError
from git.cmd import Git
from git.refs import Head, RemoteReference
from git.repo import Repo

from .exceptions import *
from .util import print_stderr

__all__ = ["GitRepoState"]

log = logging.getLogger(__name__)


def _parse_git_remote_url(url: str) -> tuple[str, str]:
    """
    Parse a git remote URL into a GitHub (account, repo) pair.

    :raises InvalidRemoteError: If the URL can't be parsed correctly.
    """
    try:
        account, repo = (
            url.split("https://github.com/")[-1]
            .split("git@github.com:")[-1]
            .split(".git")[0]
            .split("/")
        )
    except ValueError:
        raise InvalidRemoteError(f"Failed to parse GitHub repo path from remote '{url}'")
    return account, repo


def _resolve_repo() -> Repo:
    try:
        return Repo(".")
    except InvalidGitRepositoryError as e:
        raise GitError("gantry must be run from the ROOT of a valid git repository!") from e


@dataclass
class GitRepoState:
    """
    Represents the state of a local git repository.

    .. tip::
        Use :meth:`from_env()` to instantiate this class.
    """

    repo: str
    """
    The repository name, e.g. ``"allenai/beaker-gantry"``.
    """
    repo_url: str
    """
    The repository URL for cloning, e.g. ``"https://github.com/allenai/beaker-gantry"``.
    """
    ref: str
    """
    The current ref.
    """
    branch: str | None = None
    """
    The current active branch, if any.
    """

    @property
    def is_dirty(self) -> bool:
        """
        If the local repository state is dirty (uncommitted changes).
        """
        repo = _resolve_repo()
        return repo.is_dirty()

    @cached_property
    def is_public(self) -> bool:
        """
        If the repository is public.
        """
        response = requests.get(self.repo_url)
        if response.status_code not in {200, 404}:
            response.raise_for_status()
        return response.status_code == 200

    @property
    def ref_url(self) -> str:
        """
        The URL to the current :data:`ref`.
        """
        return f"{self.repo_url}/commit/{self.ref}"

    @property
    def branch_url(self) -> str | None:
        """
        The URL to the current active :data:`branch`.
        """
        if self.branch is None:
            return None
        else:
            return f"{self.repo_url}/tree/{self.branch}"

    @classmethod
    def from_env(cls, ref: str | None = None, branch: str | None = None) -> GitRepoState:
        """
        Instantiate this class from the root of a git repository.

        :raises ~gantry.exceptions.GitError: If this method isn't called from the
            root of a valid git repository.
        :raises ~gantry.exceptions.UnpushedChangesError: If there are unpushed commits.
        :raises ~gantry.exceptions.RemoteBranchNotFoundError: If the local branch is not tracking
            a remote branch.
        """
        repo = _resolve_repo()
        git = Git(".")

        git_ref = ref or str(repo.commit())
        remote = repo.remote()

        active_branch: Head | None = None
        if branch is not None:
            active_branch = Head(repo, f"refs/heads/{branch}")
        else:
            try:
                active_branch = repo.active_branch
            except TypeError:
                print_stderr(
                    "[yellow]Repo is in 'detached HEAD' state which will result in cloning the entire repo at runtime.\n"
                    "It's recommended to run gantry from a branch instead.[/]"
                )

        remote_branch: RemoteReference | None = None
        if active_branch is not None:
            remote_branch = active_branch.tracking_branch()
            if remote_branch is None:
                raise RemoteBranchNotFoundError(
                    f"Failed to resolve remote tracking branch for local branch '{active_branch.name}'.\n"
                    f"Please make sure your branch exists on the remote, e.g. 'git push --set-upstream {remote.name}'."
                )

        remote_branches_containing_ref = {
            remote_branch_name.strip()
            for remote_branch_name in cast(
                str,
                git.execute(["git", "branch", "-r", "--contains", git_ref], stdout_as_string=True),
            )
            .strip()
            .split("\n")
        }

        branch_name: str | None = None
        if remote_branch is not None:
            assert remote_branch.name.startswith(remote_branch.remote_name + "/")
            remote = repo.remote(remote_branch.remote_name)
            branch_name = remote_branch.name.replace(remote_branch.remote_name + "/", "", 1)
            if remote_branch.name not in remote_branches_containing_ref:
                raise UnpushedChangesError(
                    f"Current git ref '{git_ref}' does not appear to exist on the remote tracking branch '{remote_branch.name}'!\n"
                    "Please push your changes and try again."
                )
        else:
            if not remote_branches_containing_ref:
                raise UnpushedChangesError(
                    f"Current git ref '{git_ref}' does not appear to exist on the remote!\n"
                    "Please push your changes and try again."
                )

        account, repo_name = _parse_git_remote_url(remote.url)

        return cls(
            repo=f"{account}/{repo_name}",
            repo_url=f"https://github.com/{account}/{repo_name}",
            ref=git_ref,
            branch=branch_name,
        )



DEFAULT_IMAGE = "petew/gantry"

ENTRYPOINT = "entrypoint.sh"

GITHUB_TOKEN_SECRET = "GITHUB_TOKEN"

CONDA_ENV_FILE = "environment.yml"

CONDA_ENV_FILE_ALTERNATE = "environment.yaml"

PIP_REQUIREMENTS_FILE = "requirements.txt"

RUNTIME_DIR = "/gantry-runtime"

RESULTS_DIR = "/results"

METRICS_FILE = f"{RESULTS_DIR}/metrics.json"
